#pragma once

#ifndef OCTREE_CREATE_H
#define OCTREE_CREATE_H




#include "synchronized.h"
#include "octree_node.h"
#include "shape.h"

#include <array>
#include <algorithm>

#include <assert.h>




namespace Octree {




namespace {




/*==============================================================================
	CreateNode function
==============================================================================*/


template< int t_ColorFormat, int t_Platform >
Node< t_ColorFormat, t_Platform > CreateNode(
	Synchronized::Heap< t_Platform >& heap,
	Synchronized::TextureHeap< t_ColorFormat, t_Platform >& textureHeap,
	unsigned int const depthRemaining,
	Vector< double, 3 > const& position,
	double side,
	Shape< unsigned int > const shape
)
{
	Node< t_ColorFormat, t_Platform > result;

	Vector< double, 3 > const center = position + ( side / 2 );
	bool const split = shape.Split( position, side );
	bool const inside = ( shape.Inside( 0, center ) != 0 );

	if ( ( depthRemaining > 0 ) && split ) {

		result.children = heap.template Allocate< Node< t_ColorFormat, t_Platform > >( 8 );

		double const sideChild = side / 2;
		Vector< double, 3 > positionChild;

		unsigned int index = 0;
		positionChild[ 2 ] = position[ 2 ];
		for ( unsigned int ii = 0; ii < 2; ++ii ) {

			positionChild[ 1 ] = position[ 1 ];
			for ( unsigned int jj = 0; jj < 2; ++jj ) {

				positionChild[ 0 ] = position[ 0 ];
				for ( unsigned int kk = 0; kk < 2; ++kk ) {

					result.children.Assign( index, CreateNode( heap, textureHeap, depthRemaining - 1, positionChild, sideChild, shape ) );

					positionChild[ 0 ] += sideChild;
					++index;
				}
				positionChild[ 1 ] += sideChild;
			}
			positionChild[ 2 ] += sideChild;
		}
	}

	if ( split || inside ) {

		result.texture = textureHeap.Allocate( 2 );

		double const textureSide = ( side / 4 );
		Vector< double, 3 > texturePosition;

		unsigned int index = 0;
		texturePosition[ 2 ] = position[ 2 ] + ( textureSide / 2 );
		for ( unsigned int ii = 0; ii < 4; ++ii ) {

			texturePosition[ 1 ] = position[ 1 ] + ( textureSide / 2 );
			for ( unsigned int jj = 0; jj < 4; ++jj ) {

				texturePosition[ 0 ] = position[ 0 ] + ( textureSide / 2 );
				for ( unsigned int kk = 0; kk < 4; ++kk ) {

					std::array< uint8_t, 3 > color;
					switch( shape.Inside( 0, texturePosition ) ) {

						case 1: {

							color[ 0 ] = ( ( texturePosition[ 0 ] > 0 ) ? 0x00 : 0x80 );
							color[ 1 ] = ( ( texturePosition[ 1 ] > 0 ) ? 0x00 : 0x80 );
							color[ 2 ] = ( ( texturePosition[ 2 ] > 0 ) ? 0x00 : 0x80 );

							uint8_t const xxIndex = static_cast< uint8_t >( ( texturePosition[ 0 ] + 1 ) * 512 + 0.5 );
							uint8_t const yyIndex = static_cast< uint8_t >( ( texturePosition[ 1 ] + 1 ) * 512 + 0.5 );
							uint8_t const zzIndex = static_cast< uint8_t >( ( texturePosition[ 2 ] + 1 ) * 512 + 0.5 );

							color[ 0 ] += ( ( yyIndex ^ zzIndex ) & 0x7f );
							color[ 1 ] += ( ( xxIndex ^ zzIndex ) & 0x7f );
							color[ 2 ] += ( ( xxIndex ^ yyIndex ) & 0x7f );

							break;
						}

						case 2: {

							color[ 0 ] = static_cast< uint8_t >( std::max( 0.0, std::min( 1.0, ( texturePosition[ 0 ] + 0.1 ) / 0.2 ) ) * 0xef + 0x10 );
							color[ 1 ] = static_cast< uint8_t >( std::max( 0.0, std::min( 1.0, ( texturePosition[ 1 ] + 0.1 ) / 0.2 ) ) * 0xff + 0x00 );
							color[ 2 ] = static_cast< uint8_t >( std::max( 0.0, std::min( 1.0, ( texturePosition[ 2 ] + 0.1 ) / 0.2 ) ) * 0xef + 0x10 );

							break;
						}

						case 3: {

							color[ 0 ] = 0xc0;
							color[ 1 ] = 0xc0;
							color[ 2 ] = 0xc0;

							// shuffle the colors a bit so that we can distinguish between adjacent voxels
							color[ 0 ] ^= ( ( ii & 1 ) << 4 );
							color[ 1 ] ^= ( ( jj & 1 ) << 3 );
							color[ 2 ] ^= ( ( kk & 1 ) << 4 );

							break;
						}

						case 4: {

							color[ 0 ] = ( ( texturePosition[ 0 ] > 0 ) ? 0xc0 : 0x20 );
							color[ 1 ] = ( ( texturePosition[ 1 ] > 0 ) ? 0xc0 : 0x20 );
							color[ 2 ] = ( ( texturePosition[ 2 ] > 0 ) ? 0xc0 : 0x20 );

							// shuffle the colors a bit so that we can distinguish between adjacent voxels
							color[ 0 ] ^= ( ( ii & 1 ) << 5 );
							color[ 1 ] ^= ( ( jj & 1 ) << 4 );
							color[ 2 ] ^= ( ( kk & 1 ) << 5 );

							break;
						}

						default: {

							color[ 0 ] = 0x00;
							color[ 1 ] = 0x00;
							color[ 2 ] = 0x00;

							break;
						}
					}

					assert( index < 64 );
					result.texture.Assign( index, Color< t_ColorFormat >( color[ 0 ], color[ 1 ], color[ 2 ] ) );

					++index;
					texturePosition[ 0 ] += textureSide;
				}

				texturePosition[ 1 ] += textureSide;
			}

			texturePosition[ 2 ] += textureSide;
		}
	}

	return result;
}




/*==============================================================================
	SubtractNode function
==============================================================================*/


template< int t_ColorFormat, int t_Platform >
Node< t_ColorFormat, t_Platform > SubtractNode(
	Synchronized::Heap< t_Platform >& heap,
	Synchronized::TextureHeap< t_ColorFormat, t_Platform >& textureHeap,
	Node< t_ColorFormat, t_Platform > const& originalNode,
	unsigned int const depthRemaining,
	Vector< double, 3 > const& position,
	double side,
	Shape< unsigned int > const shape
)
{
	Node< t_ColorFormat, t_Platform > result;

	Vector< double, 3 > const center = position + ( side / 2 );
	bool const split = shape.Split( position, side );
	bool const inside = ( shape.Inside( 0, center ) != 0 );

	if ( ( depthRemaining > 0 ) && split ) {

		if ( originalNode.children ) {

			result.children = heap.template Allocate< Node< t_ColorFormat, t_Platform > >( 8 );

			double const sideChild = side / 2;
			Vector< double, 3 > positionChild;

			unsigned int index = 0;
			positionChild[ 2 ] = position[ 2 ];
			for ( unsigned int ii = 0; ii < 2; ++ii ) {

				positionChild[ 1 ] = position[ 1 ];
				for ( unsigned int jj = 0; jj < 2; ++jj ) {

					positionChild[ 0 ] = position[ 0 ];
					for ( unsigned int kk = 0; kk < 2; ++kk ) {

						result.children.Assign( index, SubtractNode( heap, textureHeap, originalNode.children[ index ], depthRemaining - 1, positionChild, sideChild, shape ) );

						positionChild[ 0 ] += sideChild;
						++index;
					}
					positionChild[ 1 ] += sideChild;
				}
				positionChild[ 2 ] += sideChild;
			}
		}
		// **TODO: if texture and no children, we should continue splitting, generating new data
	}
	else if ( ! inside )
		result.children = originalNode.children;

	if ( originalNode.texture ) {

		if ( split ) {

			result.texture = textureHeap.Allocate( 2 );

			double const textureSide = ( side / 4 );
			Vector< double, 3 > texturePosition;

			unsigned int index = 0;
			texturePosition[ 2 ] = position[ 2 ] + ( textureSide / 2 );
			for ( unsigned int ii = 0; ii < 4; ++ii ) {

				texturePosition[ 1 ] = position[ 1 ] + ( textureSide / 2 );
				for ( unsigned int jj = 0; jj < 4; ++jj ) {

					texturePosition[ 0 ] = position[ 0 ] + ( textureSide / 2 );
					for ( unsigned int kk = 0; kk < 4; ++kk ) {

						if ( ! shape.Inside( 0, texturePosition ) ) {

							assert( index < 64 );
							result.texture.Assign( index, originalNode.texture[ index ] );
						}

						++index;
						texturePosition[ 0 ] += textureSide;
					}

					texturePosition[ 1 ] += textureSide;
				}

				texturePosition[ 2 ] += textureSide;
			}
		}
		else if ( ! inside )
			result.texture = originalNode.texture;
	}

	return result;
}




}    // anonymous namespace




/*==============================================================================
	Create function
==============================================================================*/


template< int t_ColorFormat, int t_Platform >
typename Synchronized::Heap< t_Platform >::template Pointer< Node< t_ColorFormat, t_Platform > > const Create(
	Synchronized::Heap< t_Platform >& heap,
	Synchronized::TextureHeap< t_ColorFormat, t_Platform >& textureHeap,
	unsigned int const logResolution = 7
)
{
	Shape< unsigned int > shape;
	// central shape
	{ Vector< double, 3 > center = { { 0.0, 0.0, 0.0 } }; shape.PushBall( 1, center, 1.0, 0.6 ); }
	{ Vector< double, 3 > center = { { 0.0, 0.0, 0.0 } }; shape.PushBall( 0, center, 0.9, 0.6 ); }
	{ Vector< double, 3 > center = { { 0.0, 0.0, 0.0 } }; shape.PushBall( 2, center, 0.1, 2.0 ); }
	// small balls on the corners
	{ Vector< double, 3 > center = { {  0.00,  0.00,  0.97 } }; shape.PushBall( 3, center, 0.03, 2.0 ); }
	{ Vector< double, 3 > center = { {  0.00,  0.00, -0.97 } }; shape.PushBall( 3, center, 0.03, 2.0 ); }
	{ Vector< double, 3 > center = { {  0.00,  0.97,  0.00 } }; shape.PushBall( 3, center, 0.03, 2.0 ); }
	{ Vector< double, 3 > center = { {  0.00, -0.97,  0.00 } }; shape.PushBall( 3, center, 0.03, 2.0 ); }
	{ Vector< double, 3 > center = { {  0.97,  0.00,  0.00 } }; shape.PushBall( 3, center, 0.03, 2.0 ); }
	{ Vector< double, 3 > center = { { -0.97,  0.00,  0.00 } }; shape.PushBall( 3, center, 0.03, 2.0 ); }
	// squarish balls on the sides
	{ Vector< double, 3 > center = { {  0.3,  0.3,  0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { {  0.3,  0.3, -0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { {  0.3, -0.3,  0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { {  0.3, -0.3, -0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { { -0.3,  0.3,  0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { { -0.3,  0.3, -0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { { -0.3, -0.3,  0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }
	{ Vector< double, 3 > center = { { -0.3, -0.3, -0.3 } }; shape.PushBall( 4, center, 0.1, 4.0 ); }

	assert( logResolution > 0 );

	// populate the tree
	Vector< double, 3 > const position = { { -1, -1, -1 } };
	typename Synchronized::Heap< t_Platform >::template Pointer< Node< t_ColorFormat, t_Platform > > pRoot = heap.template Allocate< Node< t_ColorFormat, t_Platform > >();
	pRoot.Assign( CreateNode( heap, textureHeap, logResolution, position, 2, shape ) );

	return pRoot;
}




/*==============================================================================
	Subtract function
==============================================================================*/


template< int t_ColorFormat, int t_Platform >
typename Synchronized::Heap< t_Platform >::template Pointer< Node< t_ColorFormat, t_Platform > > const Subtract(
	Synchronized::Heap< t_Platform >& heap,
	Synchronized::TextureHeap< t_ColorFormat, t_Platform >& textureHeap,
	typename Synchronized::Heap< t_Platform >::template Pointer< Node< t_ColorFormat, t_Platform > > const& pOriginalRoot,
	Vector< double, 3 > const& center,
	double const radius,
	double const power,
	unsigned int const logResolution = 7
)
{
	Vector< double, 3 > const scaledCenter = center * 2 - 1;
	Shape< unsigned int > shape;
	shape.PushBall( 1, scaledCenter, radius, power );

	// populate the tree
	Vector< double, 3 > const position = { { -1, -1, -1 } };
	typename Synchronized::Heap< t_Platform >::template Pointer< Node< t_ColorFormat, t_Platform > > pRoot = heap.template Allocate< Node< t_ColorFormat, t_Platform > >();
	pRoot.Assign( SubtractNode( heap, textureHeap, *pOriginalRoot, logResolution, position, 2, shape ) );

	return pRoot;
}




}    // namespace Octree




#endif    // OCTREE_CREATE_H
