#pragma once

#ifndef RENDERER_STATE_OPENCL_H
#define RENDERER_STATE_OPENCL_H




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include "renderer_state_base.h"
#include "renderer_camera.h"
#include "synchronized.h"
#include "platform.h"
#include "octree.h"
#include "opencl.h"
#include "vector.h"
#include "helpers.h"
#include "settings.h"

#include <CL/cl.h>

#include <array>

#include <stdint.h>




namespace Renderer {




/*==============================================================================
	State< PLATFORM_OPENCL > specialization
==============================================================================*/


template<>
struct State< PLATFORM_OPENCL > {

	inline State( std::shared_ptr< OpenCL::State > const& pState );


	template< int t_ColorFormat >
	void TraceScreen(
		uint8_t* const colors,
		Vector< unsigned int, 2 > const dimension,                                                                        // width, height
		unsigned int const pitch,                                                                                         // bytes per row
		Synchronized::Heap< PLATFORM_OPENCL > const& heap,                                                                // heap containing the octree
		Synchronized::TextureHeap< t_ColorFormat, PLATFORM_OPENCL > const& textureHeap,                                   // texture heap containing the octree's textures
		Synchronized::Heap< PLATFORM_OPENCL >::Pointer< Octree::Node< t_ColorFormat, PLATFORM_OPENCL > > const& pRoot,    // root of the octree
		Camera const& camera                                                                                              // camera matrix (x,y,z vectors, and position)
	) const;


private:

	std::shared_ptr< OpenCL::State > m_pState;
	std::unique_ptr< _cl_kernel, OpenCL::Deleter< _cl_kernel > > m_kernel;
};




/*==============================================================================
	State< PLATFORM_OPENCL > methods
==============================================================================*/


State< PLATFORM_OPENCL >::State( std::shared_ptr< OpenCL::State > const& pState ) : m_pState( pState ) {

	cl_int result = CL_SUCCESS;

	static char const* const filenames[] = { "settings.h", "trace.cl", NULL };
	std::unique_ptr< OpenCL::Program > pProgram( new OpenCL::Program( m_pState, filenames, "-Werror -cl-nv-verbose -cl-no-signed-zeros -cl-fast-relaxed-math", "build.log" ) );

	m_kernel.reset( clCreateKernel( pProgram->BuiltProgram(), "TraceScreen", &result ) );
	if ( result != CL_SUCCESS )
		THROW_OPENCL_EXCEPTION( "clCreateKernel failed!", result );
}


template< int t_ColorFormat >
void State< PLATFORM_OPENCL >::TraceScreen(
	uint8_t* const colors,
	Vector< unsigned int, 2 > const dimension,
	unsigned int const pitch,
	Synchronized::Heap< PLATFORM_OPENCL > const& heap,
	Synchronized::TextureHeap< t_ColorFormat, PLATFORM_OPENCL > const& textureHeap,
	Synchronized::Heap< PLATFORM_OPENCL >::Pointer< Octree::Node< t_ColorFormat, PLATFORM_OPENCL > > const& pRoot,
	Camera const& camera
) const
{
	// **TODO: store colors in a 2d image object. this should save us the effort of sending dimension and pitch to the OpenCL kernel.

	Vector< double, 3 > const& xx = camera.VectorX();
	Vector< double, 3 > const& yy = camera.VectorY();
	Vector< double, 3 > const& zz = camera.VectorZ();
	Vector< double, 3 > const& position = camera.Position();
	Vector< double, 2 > const fieldOfView = dimension * ( camera.Zoom() / dimension.Maximum() );

	double const spread = 2 * ( fieldOfView / dimension ).Minimum();    // x2 because for fov=1, the screen is 2 units wide for every 1 unit of distance

	cl_int result = CL_SUCCESS;

	cl_uint const oclColorsPitch = ( dimension[ 0 ] * sizeof( cl_uint ) + 0x3f ) & ~0x3f;
	std::unique_ptr< _cl_mem, OpenCL::Deleter< _cl_mem > > oclColors( clCreateBuffer( m_pState->Context(), CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR, dimension[ 1 ] * oclColorsPitch, NULL, &result ) );
	if ( result != CL_SUCCESS )
		THROW_OPENCL_EXCEPTION( "clCreateBuffer failed!", result );

	cl_uint const oclDepthsPitch = ( dimension[ 0 ] * sizeof( cl_float ) + 0x3f ) & ~0x3f;
	std::unique_ptr< _cl_mem, OpenCL::Deleter< _cl_mem > > oclDepths( clCreateBuffer( m_pState->Context(), CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR, dimension[ 1 ] * oclDepthsPitch, NULL, &result ) );
	if ( result != CL_SUCCESS )
		THROW_OPENCL_EXCEPTION( "clCreateBuffer failed!", result );

	{	cl_mem    const oclColorsPointer = oclColors.get();
		cl_mem    const oclDepthsPointer = oclDepths.get();
		cl_uint2  const oclDimension     = { { dimension[ 0 ], dimension[ 1 ] } };
		cl_float2 const oclFieldOfView   = { { static_cast< cl_float >( fieldOfView[ 0 ] ), static_cast< cl_float >( fieldOfView[ 1 ] ) } };
		cl_float  const oclSpread        = static_cast< cl_float >( spread );
		cl_mem    const oclHeap          = OpenCL::Convert( heap );
		cl_mem    const oclTextureHeap   = OpenCL::Convert( textureHeap );
		cl_uint   const oclRoot          = OpenCL::Convert( pRoot );
		cl_float3 const oclXX            = { { static_cast< cl_float >( xx[       0 ] ), static_cast< cl_float >( xx[       1 ] ), static_cast< cl_float >( xx[       2 ] ) } };
		cl_float3 const oclYY            = { { static_cast< cl_float >( yy[       0 ] ), static_cast< cl_float >( yy[       1 ] ), static_cast< cl_float >( yy[       2 ] ) } };
		cl_float3 const oclZZ            = { { static_cast< cl_float >( zz[       0 ] ), static_cast< cl_float >( zz[       1 ] ), static_cast< cl_float >( zz[       2 ] ) } };
		cl_float3 const oclPP            = { { static_cast< cl_float >( position[ 0 ] ), static_cast< cl_float >( position[ 1 ] ), static_cast< cl_float >( position[ 2 ] ) } };

		unsigned int index = 0;
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclColorsPointer ), &oclColorsPointer ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclDepthsPointer ), &oclDepthsPointer ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclDimension     ), &oclDimension     ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclColorsPitch   ), &oclColorsPitch   ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclDepthsPitch   ), &oclDepthsPitch   ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclFieldOfView   ), &oclFieldOfView   ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclSpread        ), &oclSpread        ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclHeap          ), &oclHeap          ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclTextureHeap   ), &oclTextureHeap   ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclRoot          ), &oclRoot          ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclXX            ), &oclXX            ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclYY            ), &oclYY            ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclZZ            ), &oclZZ            ) );
		CHECK_OPENCL( clSetKernelArg( m_kernel.get(), index++, sizeof( oclPP            ), &oclPP            ) );

		size_t const localWorkSize[] = { 320 };
		size_t const workSize[] = { ( ( dimension[ 0 ] * dimension[ 1 ] + localWorkSize[ 0 ] - 1 ) / localWorkSize[ 0 ] ) * localWorkSize[ 0 ] };
		result = clEnqueueNDRangeKernel( m_pState->Queue(), m_kernel.get(), 1, NULL, workSize, localWorkSize, 0, NULL, NULL );
		if ( result != CL_SUCCESS )
			THROW_OPENCL_EXCEPTION( "clEnqueueNDRangeKernel failed!", result );
	}

	size_t const colorsSize[] = { dimension[ 0 ] * sizeof( cl_uint ), dimension[ 1 ], 1 };
	size_t const colorsOrigin[] = { 0, 0, 0 };
	result = clEnqueueReadBufferRect( m_pState->Queue(), oclColors.get(), CL_TRUE, colorsOrigin, colorsOrigin, colorsSize, colorsSize[ 0 ], 0, pitch, 0, colors, 0, NULL, NULL );
	if ( result != CL_SUCCESS )
		THROW_OPENCL_EXCEPTION( "clEnqueueReadBuffer failed!", result );
}




}    // namespace Renderer




#endif    // SETTING_OPENCL




#endif    // RENDERER_STATE_OPENCL_H
