#include "octree.h"
#include "synchronized.h"
#include "renderer.h"
#include "vector.h"
#include "exception.h"
#include "windows_exception.h"
#include "helpers.h"
#include "settings.h"

#ifdef _OPENMP
#include <omp.h>
#endif    // _OPENMP

#define NOMINMAX
#include <windows.h>
#include <windowsx.h>
#include <wrl.h>       // ComPtr

#include <d2d1.h>
#include <dwrite.h>

#include <sstream>
#include <array>
#include <cmath>

#include <stdio.h>
#include <wchar.h>
#include <assert.h>

#pragma comment( lib, "d2d1.lib"   )
#pragma comment( lib, "dwrite.lib" )




/*==============================================================================
	Global variables
==============================================================================*/

Microsoft::WRL::ComPtr< ID2D1Factory > g_pD2DFactory;
Microsoft::WRL::ComPtr< ID2D1HwndRenderTarget > g_pRenderTarget;
Microsoft::WRL::ComPtr< ID2D1SolidColorBrush > g_pTextBrush;

Microsoft::WRL::ComPtr< IDWriteFactory > g_pDWriteFactory;
Microsoft::WRL::ComPtr< IDWriteTextFormat > g_pTextFormat;

D2D1_SIZE_U g_bitmapSize = { 0, 0 };
Microsoft::WRL::ComPtr< ID2D1Bitmap > g_pBitmap;

double g_fps = 0;
double g_rps = 0;

enum MovementKeys {

	MOVEMENT_KEY_MOVE_FORWARD = 0,
	MOVEMENT_KEY_MOVE_BACK,
	MOVEMENT_KEY_MOVE_DOWN,
	MOVEMENT_KEY_MOVE_UP,
	MOVEMENT_KEY_MOVE_RIGHT,
	MOVEMENT_KEY_MOVE_LEFT,
	MOVEMENT_KEY_TURN_COUNTERCLOCKWISE,
	MOVEMENT_KEY_TURN_CLOCKWISE,
	MOVEMENT_KEY_TURN_RIGHT,
	MOVEMENT_KEY_TURN_LEFT,
	MOVEMENT_KEY_TURN_UP,
	MOVEMENT_KEY_TURN_DOWN,
	MOVEMENT_KEY_ZOOM_IN,
	MOVEMENT_KEY_ZOOM_OUT,

	MOVEMENT_KEYS_SIZE
};
bool g_movementKeys[ MOVEMENT_KEYS_SIZE ];




/*==============================================================================
	Redraw function
==============================================================================*/


template< int t_ColorFormat, int t_Platform, int t_RendererPlatform >
void Redraw(
	unsigned int const width,
	unsigned int const height,
	Synchronized::Heap< t_Platform >& heap,
	Synchronized::TextureHeap< t_ColorFormat, t_Platform >& textureHeap,
	typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const& pRoot,
	Renderer::State< t_RendererPlatform > const& renderer,
	Renderer::Camera const& camera
)
{
	if ( ( g_bitmapSize.height <= 0 ) || ( g_bitmapSize.width <= 0 ) )
		THROW_EXCEPTION( "bitmap is not initialized!" );

	BYTE* buffer = new BYTE[ g_bitmapSize.height * g_bitmapSize.width * 4 ];
	{	Vector< unsigned int, 2 > const dimensions = { { g_bitmapSize.width, g_bitmapSize.height } };

#if 0
		typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const pSubtractedRoot = Octree::Subtract(
			heap,
			textureHeap,
			pRoot,
			g_coordinates,
			0.1,
			2
		);
		heap.Synchronize();
		textureHeap.Synchronize();
#else    // 0/1
		typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const& pSubtractedRoot = pRoot;
#endif    // 0/1

		renderer.TraceScreen(
			buffer,
			dimensions,
			dimensions[ 0 ] * 4,
			heap,
			textureHeap,
			pSubtractedRoot,
			camera
		);
	}

	CHECK_HRESULT( g_pBitmap->CopyFromMemory( NULL, buffer, g_bitmapSize.width * 4 ) );

	delete[] buffer;

	wchar_t string[ 128 ];
#ifdef _OPENMP
	unsigned int const threads = omp_get_max_threads();
#else    // _OPENMP
	unsigned int const threads = 1;
#endif    // _OPENMP
	_snwprintf_s( string, ARRAYLENGTH( string ), L"threads = %u, resolution = %dx%d\n%.2f Mrays / frame\n%.2f frames / sec\n%.2f Mrays / sec", threads, g_bitmapSize.width, g_bitmapSize.height, g_bitmapSize.width * g_bitmapSize.height / 1000000.0, g_fps, g_rps / 1000000 );

	D2D1_SIZE_F const targetSize = g_pRenderTarget->GetSize();
	D2D1_RECT_F const bitmapTargetRect = { 0, 0, targetSize.width, targetSize.height };
	D2D1_RECT_F const textTargetRect   = { 8, 8, targetSize.width, targetSize.height };

	g_pRenderTarget->BeginDraw();
	g_pRenderTarget->DrawBitmap( g_pBitmap.Get(), bitmapTargetRect );
	g_pRenderTarget->DrawText(
		string,
		wcslen( string ),
		g_pTextFormat.Get(),
		&textTargetRect,
		g_pTextBrush.Get(),
		D2D1_DRAW_TEXT_OPTIONS_NONE,
		DWRITE_MEASURING_MODE_NATURAL
	);
	g_pRenderTarget->EndDraw();
}




/*==============================================================================
	WindowProc function
==============================================================================*/


LRESULT CALLBACK WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam ) {

	LRESULT result = 0;

	try {

		switch( message ) {

			case WM_KEYUP: {

				switch( wParam ) {
					case 'W':      { g_movementKeys[ MOVEMENT_KEY_MOVE_FORWARD          ] = false; break; }
					case 'S':      { g_movementKeys[ MOVEMENT_KEY_MOVE_BACK             ] = false; break; }
					case VK_NEXT:  { g_movementKeys[ MOVEMENT_KEY_MOVE_DOWN             ] = false; break; }
					case VK_PRIOR: { g_movementKeys[ MOVEMENT_KEY_MOVE_UP               ] = false; break; }
					case 'D':      { g_movementKeys[ MOVEMENT_KEY_MOVE_RIGHT            ] = false; break; }
					case 'A':      { g_movementKeys[ MOVEMENT_KEY_MOVE_LEFT             ] = false; break; }
					case 'Q':      { g_movementKeys[ MOVEMENT_KEY_TURN_COUNTERCLOCKWISE ] = false; break; }
					case 'E':      { g_movementKeys[ MOVEMENT_KEY_TURN_CLOCKWISE        ] = false; break; }
					case VK_RIGHT: { g_movementKeys[ MOVEMENT_KEY_TURN_RIGHT            ] = false; break; }
					case VK_LEFT:  { g_movementKeys[ MOVEMENT_KEY_TURN_LEFT             ] = false; break; }
					case VK_UP:    { g_movementKeys[ MOVEMENT_KEY_TURN_UP               ] = false; break; }
					case VK_DOWN:  { g_movementKeys[ MOVEMENT_KEY_TURN_DOWN             ] = false; break; }
					case VK_HOME:  { g_movementKeys[ MOVEMENT_KEY_ZOOM_OUT              ] = false; break; }
					case VK_END:   { g_movementKeys[ MOVEMENT_KEY_ZOOM_IN               ] = false; break; }
					case VK_ESCAPE: { DestroyWindow( hWnd ); break; }
				};
				break;
			}

			case WM_KEYDOWN: {

				switch( wParam ) {
					case 'W':      { g_movementKeys[ MOVEMENT_KEY_MOVE_FORWARD          ] = true; break; }
					case 'S':      { g_movementKeys[ MOVEMENT_KEY_MOVE_BACK             ] = true; break; }
					case VK_NEXT:  { g_movementKeys[ MOVEMENT_KEY_MOVE_DOWN             ] = true; break; }
					case VK_PRIOR: { g_movementKeys[ MOVEMENT_KEY_MOVE_UP               ] = true; break; }
					case 'D':      { g_movementKeys[ MOVEMENT_KEY_MOVE_RIGHT            ] = true; break; }
					case 'A':      { g_movementKeys[ MOVEMENT_KEY_MOVE_LEFT             ] = true; break; }
					case 'Q':      { g_movementKeys[ MOVEMENT_KEY_TURN_COUNTERCLOCKWISE ] = true; break; }
					case 'E':      { g_movementKeys[ MOVEMENT_KEY_TURN_CLOCKWISE        ] = true; break; }
					case VK_RIGHT: { g_movementKeys[ MOVEMENT_KEY_TURN_RIGHT            ] = true; break; }
					case VK_LEFT:  { g_movementKeys[ MOVEMENT_KEY_TURN_LEFT             ] = true; break; }
					case VK_UP:    { g_movementKeys[ MOVEMENT_KEY_TURN_UP               ] = true; break; }
					case VK_DOWN:  { g_movementKeys[ MOVEMENT_KEY_TURN_DOWN             ] = true; break; }
					case VK_HOME:  { g_movementKeys[ MOVEMENT_KEY_ZOOM_OUT              ] = true; break; }
					case VK_END:   { g_movementKeys[ MOVEMENT_KEY_ZOOM_IN               ] = true; break; }
				};
				break;
			}

			case WM_DESTROY: {

				PostQuitMessage( ERROR_SUCCESS );
				break;
			}

			case WM_SIZE: {

				RECT clientRect = { 0, 0, 0, 0 };
				GetClientRect( hWnd, &clientRect );
				{	D2D1_SIZE_U size = D2D1::SizeU( clientRect.right - clientRect.left, clientRect.bottom - clientRect.top );
					CHECK_HRESULT( g_pRenderTarget->Resize( size ) );
				}

				D2D1_SIZE_U bitmapSize = {
					clientRect.right  - clientRect.left,
					clientRect.bottom - clientRect.top
				};

				if ( ( bitmapSize.width != g_bitmapSize.width ) || ( bitmapSize.height != g_bitmapSize.height ) ) {

					g_bitmapSize = bitmapSize;

					D2D1_BITMAP_PROPERTIES properties;
					ZeroMemory( &properties, sizeof( properties ) );
					properties.pixelFormat.format    = DXGI_FORMAT_B8G8R8A8_UNORM;    // **NOTE: BGRA
					properties.pixelFormat.alphaMode = D2D1_ALPHA_MODE_IGNORE;
					properties.dpiX = 96.0;
					properties.dpiY = 96.0;

					CHECK_HRESULT( g_pRenderTarget->CreateBitmap( g_bitmapSize, properties, &g_pBitmap ) );
				}

				result = DefWindowProc( hWnd, message, wParam, lParam );
				break;
			}

			default: {

				result = DefWindowProc( hWnd, message, wParam, lParam );
				break;
			}
		}
	}
	catch( std::exception& exception ) {

#ifdef _UNICODE
		char const* const what8 = exception.what();
		wchar_t what[ 256 ];
		MultiByteToWideChar( CP_UTF8, 0, what8, -1, what, ARRAYLENGTH( what ) );
#else    // _UNICODE
		char const* const what = exception.what();
#endif    // _UNICODE

		MessageBox( NULL, what, TEXT( "WindowProc - caught exception" ), MB_ICONEXCLAMATION | MB_OK );
		PostQuitMessage( ERROR_UNHANDLED_EXCEPTION );
	}
	catch( ... ) {

		MessageBox( NULL, TEXT( "<unknown error>" ), TEXT( "WindowProc - caught exception" ), MB_ICONEXCLAMATION | MB_OK );
		PostQuitMessage( ERROR_UNHANDLED_EXCEPTION );
	}

	return result;
}




/*==============================================================================
	WinMain function
==============================================================================*/


int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow ) {

	LPCTSTR szClassName = TEXT( "WindowClass" );
	LPCTSTR szTitle = TEXT( "Voxel" );

	int result = ERROR_SUCCESS;

	try {

		// ==== initialize voxel and keyboard data ====

		//enum { COLOR_FORMAT = COLOR_FORMAT_RGB_565 };
		enum { COLOR_FORMAT = COLOR_FORMAT_RGBA_8888 };
#if ( SETTING_OPENCL != 0 )
		std::shared_ptr< OpenCL::State > pState( new OpenCL::State() );
		Synchronized::Heap< PLATFORM_OPENCL > heap( pState, ( 1u << 25 ) );
		Synchronized::TextureHeap< COLOR_FORMAT, PLATFORM_OPENCL > textureHeap( pState, 27 );
		Synchronized::Heap< PLATFORM_OPENCL >::Pointer< Octree::Node< COLOR_FORMAT, PLATFORM_OPENCL > > pRoot = Octree::Create( heap, textureHeap );
		Renderer::State< PLATFORM_OPENCL > renderer( pState );
#else    // SETTING_OPENCL
		Synchronized::Heap< PLATFORM_HOST > heap;
		Synchronized::TextureHeap< COLOR_FORMAT, PLATFORM_HOST > textureHeap;
		Synchronized::Heap< PLATFORM_HOST >::Pointer< Octree::Node< COLOR_FORMAT, PLATFORM_HOST > > pRoot = Octree::Create( heap, textureHeap );
		Renderer::State< PLATFORM_HOST > renderer;
#endif    // SETTING_OPENCL
		heap.Synchronize();
		textureHeap.Synchronize();

		Vector< double, 3 > const initialX = { { 1, 0, 0 } };
		Vector< double, 3 > const initialY = { { 0, 1, 0 } };
		Vector< double, 3 > const initialZ = { { 0, 0, 1 } };
		Vector< double, 3 > const initialPosition = { { 0.5, 0.5, 1.5 } };
		Renderer::Camera camera( initialX, initialY, initialZ, initialPosition );

		std::fill( g_movementKeys, g_movementKeys + ARRAYLENGTH( g_movementKeys ), false );

		double const fpsRate = 0.95;
		double const rateTurn = 0.8;
		double const rateZoom = 0.4;
		double const rateMove = 0.2;

		// ==== create main window ====

		WNDCLASSEX windowClass;
		ZeroMemory( &windowClass, sizeof( windowClass ) );
		windowClass.cbSize        = sizeof( windowClass );
		windowClass.style         = CS_HREDRAW | CS_VREDRAW;
		windowClass.lpfnWndProc   = WindowProc;
		windowClass.cbClsExtra    = 0;
		windowClass.cbWndExtra    = 0;
		windowClass.hInstance     = hInstance;
		windowClass.hIcon         = LoadIcon( NULL, IDI_APPLICATION );
		windowClass.hCursor       = LoadCursor( NULL, IDC_ARROW );
		windowClass.hbrBackground = ( HBRUSH )COLOR_WINDOW;
		windowClass.lpszMenuName  = NULL;
		windowClass.lpszClassName = szClassName;
		windowClass.hIconSm       = LoadIcon( NULL, IDI_APPLICATION );
		if ( ! RegisterClassEx( &windowClass ) )
			THROW_WINDOWS_EXCEPTION( "RegisterClassEx failed!", HRESULT_FROM_WIN32( GetLastError() ) );

		RECT windowRect = { 0, 0, 800, 600 };
		AdjustWindowRect(
			&windowRect,
			WS_OVERLAPPEDWINDOW,
			//WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
			FALSE
		);

		HWND hWnd = CreateWindowEx(
			WS_EX_CLIENTEDGE,
			szClassName,
			szTitle,
			WS_OVERLAPPEDWINDOW,
			//WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
			CW_USEDEFAULT,
			CW_USEDEFAULT,
			windowRect.right - windowRect.left,
			windowRect.bottom - windowRect.top,
			NULL,
			NULL,
			hInstance,
			NULL
		);
		if ( hWnd == NULL )
			THROW_EXCEPTION( "CreateWindowEx failed!" );

		// ==== initialize DirectX ====

		D2D1_FACTORY_OPTIONS options;
		ZeroMemory( &options, sizeof( options ) );

		CHECK_HRESULT( D2D1CreateFactory( D2D1_FACTORY_TYPE_SINGLE_THREADED, __uuidof( ID2D1Factory ), &options, &g_pD2DFactory ) );

		GetClientRect( hWnd, &windowRect );
		{	D2D1_SIZE_U size = D2D1::SizeU( windowRect.right - windowRect.left, windowRect.bottom - windowRect.top );
			CHECK_HRESULT( g_pD2DFactory->CreateHwndRenderTarget( D2D1::RenderTargetProperties(), D2D1::HwndRenderTargetProperties( hWnd, size ), &g_pRenderTarget ) );
		}

		CHECK_HRESULT( g_pRenderTarget->CreateSolidColorBrush(
			D2D1::ColorF( D2D1::ColorF::White, 1 ),
			&g_pTextBrush
        ) );

		CHECK_HRESULT( DWriteCreateFactory( DWRITE_FACTORY_TYPE_SHARED, __uuidof( IDWriteFactory ), &g_pDWriteFactory ) );

		CHECK_HRESULT( g_pDWriteFactory->CreateTextFormat(
			L"Verdana",
			NULL,
			DWRITE_FONT_WEIGHT_REGULAR,
			DWRITE_FONT_STYLE_NORMAL,
			DWRITE_FONT_STRETCH_NORMAL,
			12,
			L"en-us",
			&g_pTextFormat
		) );

		// ==== message loop ====

		ShowWindow( hWnd, nCmdShow );
		UpdateWindow( hWnd );

		double frames  = 0;
		double rays    = 0;
		double seconds = 0;
		DWORD ticks = GetTickCount();

		MSG msg;
		for ( bool done = false; ! done; ) {

			// ==== draw frame ====

			{	RECT clientRect = windowRect;
				GetClientRect( hWnd, &clientRect );

				Redraw( clientRect.right - clientRect.left, clientRect.bottom - clientRect.top, heap, textureHeap, pRoot, renderer, camera );
				ValidateRect( hWnd, NULL );
			}

			// ==== update counters ====

			DWORD const ticksOld = ticks;
			ticks = GetTickCount();
			double const deltaSeconds = static_cast< double >( ticks - ticksOld ) / 1000;

			frames = fpsRate * frames + ( 1 - fpsRate );
			rays = fpsRate * rays + ( 1 - fpsRate ) * g_bitmapSize.width * g_bitmapSize.height;
			seconds = fpsRate * seconds + ( 1 - fpsRate ) * deltaSeconds;

			g_fps = frames / seconds;
			g_rps = rays   / seconds;

			// ==== message loop ====

			for ( ; ; ) {

				BOOL peekResult = PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );
				if ( peekResult < 0 )
					THROW_EXCEPTION( "PeekMessage failed!" );
				if ( peekResult == 0 )
					break;

				TranslateMessage( &msg );
				DispatchMessage( &msg );

				if ( msg.message == WM_QUIT ) {

					result = msg.wParam;
					done = true;
					break;
				}
			}

			// ==== update position ====

			{	double const move = ( g_movementKeys[ MOVEMENT_KEY_MOVE_FORWARD ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_MOVE_BACK ] ? 1.0 : 0.0 );
				if ( move != 0 ) {

					double const rate = move * rateMove * deltaSeconds;
					camera.MoveForward( rate );
				}
			}
			{	double const move = ( g_movementKeys[ MOVEMENT_KEY_MOVE_DOWN ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_MOVE_UP ] ? 1.0 : 0.0 );
				if ( move != 0 ) {

					double const rate = move * rateMove * deltaSeconds;
					camera.MoveDown( rate );
				}
			}
			{	double const move = ( g_movementKeys[ MOVEMENT_KEY_MOVE_RIGHT ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_MOVE_LEFT ] ? 1.0 : 0.0 );
				if ( move != 0 ) {

					double const rate = move * rateMove * deltaSeconds;
					camera.MoveRight( rate );
				}
			}
			{	double const turn = ( g_movementKeys[ MOVEMENT_KEY_TURN_CLOCKWISE ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_TURN_COUNTERCLOCKWISE ] ? 1.0 : 0.0 );
				if ( turn != 0 ) {

					double const rate = turn * rateTurn * deltaSeconds;
					camera.TurnClockwise( rate );
				}
			}
			{	double const turn = ( g_movementKeys[ MOVEMENT_KEY_TURN_RIGHT ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_TURN_LEFT ] ? 1.0 : 0.0 );
				if ( turn != 0 ) {

					double const rate = turn * rateTurn * deltaSeconds;
					camera.TurnRight( rate );
				}
			}
			{	double const turn = ( g_movementKeys[ MOVEMENT_KEY_TURN_DOWN ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_TURN_UP ] ? 1.0 : 0.0 );
				if ( turn != 0 ) {

					double const rate = turn * rateTurn * deltaSeconds;
					camera.TurnDown( rate );
				}
			}
			{	double const zoom = ( g_movementKeys[ MOVEMENT_KEY_ZOOM_OUT ] ? 1.0 : 0.0 ) - ( g_movementKeys[ MOVEMENT_KEY_ZOOM_IN ] ? 1.0 : 0.0 );
				if ( zoom != 0 ) {

					double const rate = zoom * rateZoom * deltaSeconds;
					camera.ZoomOut( rate );
				}
			}
		}
	}
	catch( std::exception& exception ) {

#ifdef _UNICODE
		char const* const what8 = exception.what();
		wchar_t what[ 256 ];
		MultiByteToWideChar( CP_UTF8, 0, what8, -1, what, ARRAYLENGTH( what ) );
#else    // _UNICODE
		char const* const what = exception.what();
#endif    // _UNICODE

		MessageBox( NULL, what, TEXT( "WinMain - caught exception" ), MB_ICONEXCLAMATION | MB_OK );
		result = ERROR_UNHANDLED_EXCEPTION;
	}
	catch( ... ) {

		MessageBox( NULL, TEXT( "<unknown error>" ), TEXT( "WinMain - caught exception" ), MB_ICONEXCLAMATION | MB_OK );
		result = ERROR_UNHANDLED_EXCEPTION;
	}

	return result;
}
